/* passed to app.add_stylesheet() in conf.py to override
   readthedocs.org table width restrictions...
   from https://github.com/snide/sphinx_rtd_theme/issues/117 */
.wy-table-responsive table td, .wy-table-responsive table th {
    /* !important prevents the common CSS stylesheets from
       overriding this as on RTD they are loaded after this stylesheet */
    white-space: normal !important;
}

.wy-table-responsive {
    overflow: visible !important;
}
